/*
from	Shouxiang Wu <wusx@bu.edu>
to	Tullio Grassi <tullio.grassi@gmail.com>
date	Wed, May 21, 2008 at 6:40 PM
subject	CRC code

Hi Tullio,
Attached is the CRC code for 16bit data word.
You can apply reset at header word time slot and let it calculate
all data words until and including the last word before the 16-bit word count.
The CRC is stored in the place of this 16-bit word count.
*/

module CRC_manager   (
 input  clk, 
 input  [17:0] DIn,
 output [17:0] DOut  );

reg  HEADER_d1, HEADER_d2, HEADER_d3, TRAILER_d1;
reg  [17:0] DIn_d1, DIn_d2, DIn_d3, DO;
wire [15:0] CRC;

/* HEADER_d1 is aligned with the header of DIn_d1,
   HEADER_d3 is aligned with the header of DIn_d3.
  TRAILER_d1 is aligned with the trailer of DIn_d1 
  and with Pre-trailer2 (= word count) of DIn_d3   */
always @(posedge clk) begin 
  HEADER_d1  <= (DIn[17:16] == 2'b11);
  HEADER_d2  <= HEADER_d1;
  HEADER_d3  <= HEADER_d2;
  TRAILER_d1 <= (DIn[17:16] == 2'b01);
  DIn_d1     <= DIn;
  DIn_d2     <= DIn_d1;
  DIn_d3     <= DIn_d2;
  if (TRAILER_d1 == 1'b1) begin
    DO <= {DIn_d3[17:16], CRC};
  end else begin 
    DO <= DIn_d3;
  end	 
end 
assign DOut = DO;
CRC16D16 CRC16D16 ( .clk(clk), .init(HEADER_d3), .DIn(DIn_d3[15:0]), .CRC(CRC));

//mux18    CRC_insertion
//  ( .ck(clk),         .sel(TRAILER_d1), 
//    .Din_One ( {DIn_d3[17:16], CRC}),
//    .Din_Zero( {DIn_d3}), 
//    .Mux_Out ( {DOut} )	 );

endmodule 