-- this module is basically an input fifo for event data from pci1 or pci2 interface
-- pci_ctrl(0) trailer flag
-- pci_ctrl(1) monitor data flag
-- pci_ctrl(2) valid lower word
-- pci_ctrl(3) flags lrb_enable on pci_data(8 dwnto 0)
-- pci_ctrl(4) is reset 
-- pci_ctrl(5) is htr_buf_avl output, high indicates to pci interface that at least one
-- more HTR data block(~300 words) could be buffered
-- pci_ctrl(6) is start signal for reading LRB monitoring data
-- pci_ctrl(7) is not used
-- if dav is high, fifo_out data are valid
-- out_gnt is fifo output enable. It is a pulse of one clock duration
library IEEE,mylib;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use mylib.vtx_pack.all;
use mylib.dcc_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity lrb_if is
    Port ( clk : in std_logic;
           reset : in std_logic;
           ttc_reset : in std_logic;
			  n_strobe : in std_logic;
			  run : in std_logic;
			  htr_buf_full : in std_logic_vector(8 downto 0);
-- lrb fifo read out
           dav : out std_logic;
           out_gnt : in std_logic;
           out_rqst : out std_logic;
-- lrb_out(35) = '1' flags data truncated
-- lrb_out(34) unused
-- lrb_out(33) = '1' flags header
-- lrb_out(32) = '1' flags trailer
           lrb_out : out std_logic_vector(35 downto 0);
-- lrb monitor data
           auto_monitor : in std_logic;
           monitor_start: in STD_LOGIC;
           lrb_mon_ready: out STD_LOGIC;
-- read out control
           rd_lrb_mon: in STD_LOGIC;
           lrb_en: out STD_LOGIC_VECTOR(8 downto 0);
           addr: in STD_LOGIC_VECTOR(13 downto 0);
           r_data: out STD_LOGIC_VECTOR(31 downto 0);
-- LRB pads
			  lrb_pads : inout lrb_pad);
end lrb_if;

architecture lrb_arch of lrb_if is
signal strobe: std_logic := '0';
signal lrb_en_i: std_logic_vector(8 downto 0) := (others => '0');
signal fifo_out: std_logic_vector(35 downto 0) := (others => '0');
signal pci_data: std_logic_vector(15 downto 0) := (others => '0');
signal pci_ctrl: std_logic_vector(3 downto 0) := (others => '0');
signal buf_avl: std_logic := '0';
signal sr_strobe: std_logic := '0';
signal sr: std_logic_vector(8 downto 0) := (others => '0');
signal htr_buf_fullp: std_logic_vector(8 downto 0) := (others => '0');
signal sr_cntr: std_logic_vector(3 downto 0) := (others => '0');
signal fifo_in: std_logic_vector(35 downto 0) := (others => '0');
signal fifo_wa: std_logic_vector(10 downto 0) := (others => '0');
signal n_header: std_logic := '0';
signal rst_output_wc: std_logic := '0';
signal lrb_off: std_logic_vector(2 downto 0) := (others => '0');
signal lrb_mask: std_logic_vector(2 downto 0) := (others => '0');
signal sel_lrb: std_logic_vector(2 downto 0) := (others => '0');
signal d_mon_wa: std_logic_vector(4 downto 0) := (others => '0');
signal ld_mon_wa: std_logic := '0';
signal fifo_ra: std_logic_vector(10 downto 0) := (others => '0');
signal word_cnt: std_logic_vector(10 downto 0) := (others => '0');
signal wc_ovfl: std_logic := '0';
signal wr_fifo: std_logic := '0';
signal ec_fifo_wa: std_logic := '0';
signal rd_fifo: std_logic := '0';
signal up_cnt: std_logic := '0';
signal down_cnt: std_logic := '0';
signal blk_cnt: std_logic_vector(3 downto 0) := (others => '0');
signal last_word: std_logic := '0';
signal dav_i: std_logic := '0';
signal got_block: std_logic := '0';
signal got_all: std_logic := '0';
signal got_all_dl: std_logic := '0';
signal auto_start: std_logic := '0';
signal got_header: std_logic_vector(8 downto 0) := (others => '0');
signal lrb_mon: std_logic_vector(31 downto 0) := (others => '0');
signal lrb_mon_wa: std_logic_vector(9 downto 0) := (others => '0');
signal lrb_mon_ra: std_logic_vector(8 downto 0) := (others => '0');
signal lrb_mon_wc: std_logic_vector(3 downto 0) := (others => '0');
signal wr_lrb_mon: std_logic := '0';
signal lrb_mon_ready_i: std_logic := '0';
begin
dav <= dav_i;
lrb_en <= lrb_en_i;
lrb_out <= fifo_out;
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
  end if;
end process;
-- scramble LRB channels
htr_buf_fullp(8) <= htr_buf_full(6);
htr_buf_fullp(7) <= htr_buf_full(7);
htr_buf_fullp(6) <= htr_buf_full(8);
htr_buf_fullp(5) <= htr_buf_full(3);
htr_buf_fullp(4) <= htr_buf_full(4);
htr_buf_fullp(3) <= htr_buf_full(5);
htr_buf_fullp(2) <= htr_buf_full(0);
htr_buf_fullp(1) <= htr_buf_full(1);
htr_buf_fullp(0) <= htr_buf_full(2);
process(clk)
begin
  if(clk'event and clk = '1')then
	 if(strobe = '1')then
-- no LRB read when not in run mode or when there are not
-- enough space for one more event in the fifo
		buf_avl <= run and not blk_cnt(3) and not word_cnt(10);
		if(buf_avl = '0')then
		  sr <= "000000000";
      elsif(sr_cntr(3) = '1')then
		  sr <= lrb_en_i and not htr_buf_fullp;
		else
		  sr <= sr(7 downto 0) & '0';
		end if;
      if(sr_cntr(3) = '1')then
		  sr_cntr <= "0000";
		else
		  sr_cntr <= sr_cntr + 1;
		end if; 
		sr_strobe <= sr_cntr(3);
		lrb_pads.ctrl(4) <= reset or ttc_reset; 
		lrb_pads.ctrl(5) <= sr_strobe;
		lrb_pads.ctrl(6) <= monitor_start or auto_start;
      lrb_pads.ctrl(7) <= sr(8);
-- lrb_en_i is from Altera chip, this ensures they always agree
	   if(pci_ctrl(3) = '1')then
		  lrb_en_i <= pci_data(8 downto 0);
		end if;
	   pci_data <= lrb_pads.data;
		pci_ctrl <= lrb_pads.ctrl(3 downto 0);
    end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   n_header <= '0';
      fifo_in(35) <= '0';
	 elsif(strobe = '1')then
	   n_header <= not(pci_ctrl(0) and pci_ctrl(2)) and (n_header or (pci_ctrl(2) and not pci_ctrl(1)));
-- fifo_out(35) = '1' truncate flag
      if(pci_ctrl(2) = '1' and pci_ctrl(1) = '0')then
        fifo_in(35) <= wc_ovfl and (fifo_in(35) or not pci_ctrl(0));
		end if;
    end if;
  end if;
end process;
-- do not accept more than 512 32-bit words
i_input_wc: l_cntr
  generic map(N => 9, ld_plus_one => true)
  port map(
    d => "000000000",
    load => fifo_in(33),
    reset => reset,
    set => '0',
    ec => ec_fifo_wa,
-- header forces wc_ovfl to '0'
    en_cnt => n_header,
    clk => clk,
	 carry => wc_ovfl);
fifo_in(34) <= '0';
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
-- do not write monitor data to fifo or when 511 words reached unless trailer
      wr_fifo <= pci_ctrl(2) and not pci_ctrl(1) and (not wc_ovfl or pci_ctrl(0));
-- fifo_out(33) = '1' flags header
      fifo_in(33) <= not n_header;
-- fifo_out(32) = '1' flags trailer
	   fifo_in(32) <= pci_ctrl(0);
      fifo_in(15 downto 0) <= pci_data(15 downto 0);
	 end if;
  end if;
end process;
fifo_in(31 downto 16) <= pci_data;
g_fifo: for i in 0 to 3 generate
fifo: ramb16_s9_s9
	  generic map(SIM_COLLISION_CHECK => "NONE")
     port map(
         dia => fifo_in(i*8+7 downto i*8), 
         dipa => fifo_in(i+32 downto i+32), 
         addra => fifo_wa,
         clka => clk, 
         ssra => '0', 
         ena => strobe, 
         wea => wr_fifo,
         dib => x"00",
         dipb => "0",
         addrb => fifo_ra,
         clkb => clk,
         ssrb => '0',
         enb => '1',
         web => '0',
         dob => fifo_out(i*8+7 downto i*8),
         dopb => fifo_out(i+32 downto i+32));
end generate;
i_fifo_wa: l_cntr
  generic map(N => 11)
  port map(
    d => "00000000000",
    load => '0',
    reset => reset,
    set => '0',
    ec => ec_fifo_wa,
    en_cnt => '1',
    clk => clk,
	 q => fifo_wa);
i_fifo_ra: l_cntr
  generic map(N => 11)
  port map(
    d => "00000000000",
    load => '0',
    reset => reset,
    set => '0',
    ec => rd_fifo,
    en_cnt => '1',
    clk => clk,
	 q => fifo_ra);
i_word_cnt: updown_cntr
  generic map(N => 11)
  port map(
    rst => reset,
    set => '0',
    ce => '1',
    up => ec_fifo_wa,
    down => dav_i,
    clk => clk,
	 q => word_cnt);
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   ec_fifo_wa <= '0';
	   up_cnt <= '0';
	   down_cnt <= '0';
	 else
	   ec_fifo_wa <= n_strobe and wr_fifo;
	   up_cnt <= strobe and pci_ctrl(2) and not pci_ctrl(1) and pci_ctrl(0);
	   down_cnt <= fifo_out(32) and dav_i;
	 end if;
  end if;
end process;
i_blk_cnt: updown_cntr
  generic map(N => 4)
  port map(
    rst => reset,
    set => '0',
    ce => '1',
    up => up_cnt,
    down => down_cnt,
    clk => clk,
	 q => blk_cnt);
I_output_wc: l_cntr
  generic map(N=> 6)
  port map(
    d => "000000",
    load => '0',
    reset => rst_output_wc,
    set => '0',
    ec => '1',
    en_cnt => '1',
    clk => clk,
    carry => last_word
  );
rst_output_wc <= not dav_i;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   out_rqst <= '0';
	   got_block <= '0';
	   dav_i <= '0';
	 else
	   if(blk_cnt = "0000" and got_block = '0')then
		  out_rqst <= '0';
		else
		  out_rqst <= '1';
		end if;
		if(word_cnt(10 downto 6) = "00000")then
		  got_block <= '0';
		else
		  got_block <= '1';
		end if;
-- start output at out_gnt, end when trailer seen or 64 words reached
	   dav_i <= rd_fifo;
	 end if;
  end if;
end process;
rd_fifo <= out_gnt or (dav_i and not (fifo_out(32) or last_word));
-- LRB monitor data automatic read_out
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
		got_header <= (others => '0');
	 elsif(strobe = '1')then
	   if(auto_start = '1')then
		  got_header <= (others => '0');
		elsif(pci_ctrl(2 downto 1) = "10" and n_header = '0' and pci_data(15) = '1')then
		  case pci_data(11 downto 8) is
		    when "0000" => got_header(0) <= '1';
		    when "0001" => got_header(1) <= '1';
		    when "0010" => got_header(2) <= '1';
		    when "0100" => got_header(3) <= '1';
		    when "0101" => got_header(4) <= '1';
		    when "0110" => got_header(5) <= '1';
		    when "1000" => got_header(6) <= '1';
		    when "1001" => got_header(7) <= '1';
		    when "1010" => got_header(8) <= '1';
		    when others => null;
		  end case;
		end if;
    end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
-- if lrb_en_i = "000000000", this prevents auto_start being generated
	   got_all <= '1';
	   got_all_dl <= '1';
	   auto_start <= '0';
	 elsif(strobe = '1')then
	   if(got_header = lrb_en_i)then
		  got_all <= '1';
		else
		  got_all <= '0';
		end if;
		got_all_dl <= got_all;
		auto_start <= got_all and not got_all_dl and auto_monitor;
	 end if;
  end if;
end process;
--------------------------------------------------------------------------
-- LRB monitor data readout logic
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
		lrb_mon_ready_i <= '1';
		lrb_mon_wa(0) <= '0';
	 elsif(strobe = '1')then
		lrb_mon_ready_i <= (lrb_mon_ready_i and not (monitor_start or auto_start)) or (lrb_mask(2) and lrb_mask(1) and lrb_mask(0));
		lrb_mon_wa(0) <= pci_ctrl(2) and pci_ctrl(1);
	 end if;
  end if;
end process;
-- monitor write address generation
process(clk)
begin
  if(clk'event and clk = '1')then
	 if(strobe = '1')then
      lrb_off(2) <= not(lrb_en_i(8) or lrb_en_i(7) or lrb_en_i(6));
      lrb_off(1) <= not(lrb_en_i(5) or lrb_en_i(4) or lrb_en_i(3));
      lrb_off(0) <= not(lrb_en_i(2) or lrb_en_i(1) or lrb_en_i(0));
      ld_mon_wa <= (lrb_mon_wc(3) and lrb_mon_wc(0) and lrb_mon_wa(0)) or lrb_mon_ready_i;
	   if(lrb_mon_ready_i = '1')then
		  lrb_mask <= lrb_off;
	   elsif(lrb_mon_wc(3) = '1' and lrb_mon_wc(0) ='1' and lrb_mon_wa(0) = '1')then
		  lrb_mask <= lrb_mask or sel_lrb;
		end if;
	 end if;
  end if;
end process;
sel_lrb(0) <= not lrb_mask(0); 
sel_lrb(1) <= lrb_mask(0) and not lrb_mask(1); 
sel_lrb(2) <= lrb_mask(0) and lrb_mask(1) and not lrb_mask(2); 
I_lrb_mon_wc: l_cntr
  generic map(N=> 4)
  port map(
    d => x"0",
    load => ld_mon_wa,
    reset => '0',
    set => '0',
    ec => strobe,
    en_cnt => lrb_mon_wa(0),
    clk => clk,
    q => lrb_mon_wc
  );
d_mon_wa <= sel_lrb(2) & sel_lrb(1) & sel_lrb(2) & sel_lrb(1) & '0';
I_lrb_mon_wa: l_cntr
  generic map(N=> 5)
  port map(
    d => d_mon_wa,
    load => ld_mon_wa,
    reset => '0',
    set => '0',
    ec => strobe,
    en_cnt => lrb_mon_wa(0),
    clk => clk,
    q => lrb_mon_wa(5 downto 1)
  );
wr_lrb_mon <= (pci_ctrl(2) and pci_ctrl(1)) or lrb_mon_wa(0);
lrb_mon_ready <= lrb_mon_ready_i;
lrb_mon_wa(9 downto 6) <= "0000";
lrb_mon_ra <= "0000" & addr(4 downto 0);
I_lrb_mon: RAMB16_S18_S36 
	  generic map(SIM_COLLISION_CHECK => "NONE")
		port map (    ADDRA => lrb_mon_wa, 
			           ADDRB => lrb_mon_ra, -- PCI3 address 
              		  DIA   => pci_data,  
              		  DIPA  => "00",
              		  DIB   => x"00000000",
              		  DIPB  => x"0",
              		  WEA   => wr_lrb_mon,    
              		  WEB   => '0',      
              		  CLKA  => clk,  
              		  CLKB  => clk,    
              		  SSRA  => '0', 
              		  SSRB  => '0', 
              		  ENA   => strobe, 
              		  ENB   => strobe, 
              		  DOB   => lrb_mon); -- output 						               		  
r_data <= lrb_mon when rd_lrb_mon = '1' else (others => 'Z');
end lrb_arch;
