library IEEE,mylib;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use mylib.vtx_pack.all;
use mylib.dcc_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity monitor is
    port (
-- reset input, should be connected to the inverted LOCK output of dll
        reset: in STD_LOGIC;
-- system clock input
        clk: in STD_LOGIC;
-- download data write
        run: in STD_LOGIC;
        w_data: in STD_LOGIC_VECTOR (31 downto 0);
        r_data: out STD_LOGIC_VECTOR (31 downto 0);
        n_strobe: in STD_LOGIC;
-- monitor event scaling factor
        wr_scale: in STD_LOGIC;
        rd_scale: in STD_LOGIC;
-- monitor event word count
        mon_buf_empty: out STD_LOGIC;
        mon_buf_full: out STD_LOGIC;
        mon_buf_afull: out STD_LOGIC;
        inc_ddr_pa: in STD_LOGIC;
        rd_mon_wc: in STD_LOGIC;
-- monitored event data
        mon_re: in STD_LOGIC;
        mon_avl: out STD_LOGIC;
        mon_we: out STD_LOGIC;
        mon_data: out STD_LOGIC_VECTOR (33 downto 0);
        mon_wp: out STD_LOGIC_VECTOR (6 downto 0);
        ddr_pa: in STD_LOGIC_VECTOR (6 downto 0);
        inc_mon_cntr: out STD_LOGIC;
        catch_offset: out STD_LOGIC;
-- data in write enable
        slink_we: in STD_LOGIC;
-- header and trailer are required to be marked as control data
        slink_data: in STD_LOGIC_VECTOR (35 downto 0)
    );
end monitor;

architecture my_arch of monitor is
signal strobe: std_logic := '0';
signal scale: std_logic_vector(31 downto 0) := x"00000000";
signal not_scale: std_logic_vector(15 downto 0) := (others => '0');
signal mon_wc: std_logic_vector(15 downto 0) := (others => '0');
signal ce_wc_reg_wa: std_logic := '0';
signal ec_fifo_ra: std_logic := '0';
signal mon_we_i: std_logic := '0';
signal mon_re_dl: std_logic := '0';
signal down_count: std_logic := '0';
signal evt_cnt: std_logic_vector(7 downto 0) := (others => '0');
signal we_fifo: std_logic := '0';
signal fifo_wa: std_logic_vector(8 downto 0) := (others => '0');
signal fifo_ra: std_logic_vector(8 downto 0) := (others => '0');
signal wc_reg_wa: std_logic_vector(9 downto 0) := (others => '0');
signal wc_reg_ra: std_logic_vector(9 downto 0) := (others => '0');
signal wc_reg: std_logic_vector(31 downto 0) := (others => '0');
signal fifo_din: std_logic_vector(35 downto 0) := (others => '0');
signal fifo_dout: std_logic_vector(35 downto 0) := (others => '0');
signal ce_scale: std_logic := '0';
signal ld_scale: std_logic := '0';
signal up_count: std_logic := '0';
signal mon_enp: std_logic := '0';
signal mon_enp2: std_logic := '0';
signal mon_en: std_logic := '0';
signal scale_carry: std_logic := '0';
signal pending: std_logic := '0';
signal catch_empty: std_logic := '0';
signal buf_empty: std_logic := '0';
signal mon_buf_empty_i: std_logic := '0';
-- error catching signals
signal htr_summaryp: std_logic := '0';
signal htr_summary: std_logic := '0';
signal htr_summary_q: std_logic := '0';
signal ec_catch_cntr: std_logic := '0';
signal catch_enable: std_logic := '0';
signal catch_offset_i: std_logic := '0';
signal catch_wc_reg_ra7: std_logic := '0';
signal catch_cntr: std_logic_vector(6 downto 0) := (others => '0');
signal catch_trigger: std_logic_vector(7 downto 0) := (others => '0');
-------------
begin
catch_offset <= catch_offset_i;
mon_buf_empty <= mon_buf_empty_i;
--mon_buf_full <= evt_cnt(7);
--mon_buf_afull <= evt_cnt(6);
mon_we <= mon_we_i;
mon_data <= fifo_dout(33 downto 0);
mon_wp <= wc_reg_wa(6 downto 0);
wc_reg_ra(6 downto 0) <= ddr_pa;
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
	 fifo_din <= slink_data;
	 we_fifo <= slink_we;
  end if;
end process;
-- monitor event fifo
mon_fifo: RAMB16_S36_S36
	  generic map(SIM_COLLISION_CHECK => "NONE")
  port map(DIA => fifo_din(31 downto 0),
  			  DIPA => fifo_din(35 downto 32),
  			  DIB => x"00000000",
  			  DIPB => x"0",
  			  ENA => mon_en,
  			  ENB => '1',
  			  WEA => we_fifo,
  			  WEB => '0',
  			  SSRA => '0',
  			  SSRB => '0',
			  CLKA => clk,
			  CLKB => clk,
			  ADDRA => fifo_wa,
			  ADDRB => fifo_ra,
			  DOB => fifo_dout(31 downto 0),
			  DOPB => fifo_dout(35 downto 32)
			  );
---------------------------------		
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   fifo_wa <= (others => '0');
	   fifo_ra <= (others => '0');
	 else
	   if(we_fifo = '1' and mon_en = '1')then
		  if(fifo_din(32) = '1')then
		    fifo_wa(8 downto 6) <= fifo_wa(8 downto 6) + 1;
		    fifo_wa(5 downto 0) <= (others => '0');
		  else
		    fifo_wa <= fifo_wa + 1;
		  end if;
		end if; 
	   if(ec_fifo_ra = '1')then
		  if(fifo_dout(32) = '1')then
		    fifo_ra(8 downto 6) <= fifo_ra(8 downto 6) + 1;
		    fifo_ra(5 downto 0) <= (others => '0');
		  else
		    fifo_ra <= fifo_ra + 1;
		  end if;
		end if;
    end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   mon_avl <= '0';
      ec_fifo_ra <= '0';
	   mon_re_dl <= '0';
	   mon_we_i <= '0';
    else
      if(ec_fifo_ra = '1' or fifo_wa(8 downto 6) = fifo_ra(8 downto 6))then
        mon_avl <= '0';
		else
        mon_avl <= '1';
      end if;
      if(mon_re = '1')then
        ec_fifo_ra <= '1';
      elsif(fifo_ra(5 downto 0) = "111111" or fifo_dout(32) = '1')then
        ec_fifo_ra <= '0';
      end if;
	   mon_re_dl <= mon_re;
	   mon_we_i <= mon_re_dl or (mon_we_i and not fifo_dout(32) and ec_fifo_ra);
    end if;
  end if;
end process;
--------------------------------------------------------------
i_mon_wc: l_cntr
  generic map(N => 13, ld_plus_one => true)
  port map(
    d => "0000000000000",
-- load to one when header is seen
    load => fifo_dout(33),
    reset => '0',
    set => '0',
    ec => mon_we_i,
    en_cnt => '1',
    clk => clk,
	 q => mon_wc(12 downto 0));
mon_wc(15 downto 13) <= "000";
i_wc_reg: RAMB16_S18_S18
	  generic map(SIM_COLLISION_CHECK => "NONE")
  port map(DIA => mon_wc,
  			  DIPA => "00",
  			  DIB => x"0000",
  			  DIPB => "00",
  			  ENA => '1',
  			  ENB => strobe,
  			  WEA => ce_wc_reg_wa,
  			  WEB => '0',
  			  SSRA => '0',
  			  SSRB => mon_buf_empty_i,
			  CLKA => clk,
			  CLKB => clk,
			  ADDRA => wc_reg_wa,
			  ADDRB => wc_reg_ra,
			  DOB => wc_reg(15 downto 0));
wc_reg(31 downto 16) <= x"0000";
wc_reg_wa(9 downto 8)	<= "00";
wc_reg_ra(9 downto 8)	<= "00";
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   wc_reg_wa(7 downto 0) <= "00000000";
		wc_reg_ra(7) <= '0';
	 else
	   if(ce_wc_reg_wa = '1')then
		  wc_reg_wa(7 downto 0) <= wc_reg_wa(7 downto 0) + 1;
		end if;
		if(catch_wc_reg_ra7 = '1')then
		  wc_reg_ra(7) <= not wc_reg_wa(7);
		elsif(strobe = '1' and inc_ddr_pa = '1' and ddr_pa = "1111111")then
		  wc_reg_ra(7) <= not wc_reg_ra(7);
		end if;
    end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
	 if(catch_enable = '1' and catch_cntr = "0000000")then
	   catch_empty <= '1';
    else
	   catch_empty <= '0';
	 end if;
	 catch_wc_reg_ra7 <= ce_wc_reg_wa and catch_offset_i;
	 if(wc_reg_wa = wc_reg_ra)then
	   buf_empty <= '1';
    else
	   buf_empty <= '0';
	 end if;
	 if(catch_empty = '1' or buf_empty = '1')then
	   mon_buf_empty_i <= '1';
    else
	   mon_buf_empty_i <= '0';
	 end if;
  end if;
end process;
i_evt_cnt: updown_cntr
  generic map(N => 8)
  port map(
    rst => reset,
    set => '0',
    ce => '1',
    up => up_count,
    down => down_count,
    clk => clk,
	 q => evt_cnt);
down_count <= '1' when strobe = '1' and evt_cnt /= x"00" and inc_ddr_pa = '1' else '0';
r_data <= wc_reg when rd_mon_wc = '1' else (others => 'Z');
------------------------------------------------------------------------
---- monitor event scaling
------------------------------------------------------------------------
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
		if(wr_scale = '1')then
        scale(23 downto 0) <= w_data(23 downto 0);
		end if;
		if(run = '1' and scale(23 downto 16) /= x"00")then
		  catch_enable <= '1'; 
		else
		  catch_enable <= '0'; 
		end if;
		if(reset = '1' or catch_enable = '0' or wr_scale = '1')then
		  catch_offset_i <= '0';
		elsif(evt_cnt(7) = '1')then
		  catch_offset_i <= '1';
		end if;
		if(catch_enable = '0')then
		  mon_buf_full <= evt_cnt(7);
		  mon_buf_afull <= evt_cnt(6);
		else
		  mon_buf_full <= catch_cntr(6);
		  mon_buf_afull <= '0';
		end if;
	 end if;
	 if(wr_scale = '1' and strobe = '1')then
		scale(31 downto 24) <= x"00";
	 elsif(catch_cntr(6) = '0')then
		scale(31 downto 24) <= scale(31 downto 24) or (scale(23 downto 16) and catch_trigger);
	 end if;
	 if(wr_scale = '1' and strobe = '1')then
		catch_cntr <= "0000000";
	 elsif(scale(31 downto 24) /= x"00" and ec_catch_cntr = '1')then
		catch_cntr <= catch_cntr + 1;
	 end if;
	 htr_summary_q <= htr_summary;
	 ec_catch_cntr <= htr_summary_q and not htr_summary and not catch_cntr(6);
	 if(reset = '1')then
		htr_summaryp <= '0';
	 elsif(fifo_din(33) = '1' and we_fifo = '1')then
	   htr_summaryp <= '1';
	 elsif(htr_summary = '1')then
		htr_summaryp <= '0'; 
	 end if;
	 if(htr_summaryp = '1' and fifo_wa(2 downto 0) = "101")then
	   htr_summary <= '1';
	 elsif(fifo_wa(3 downto 0) = x"4")then
	   htr_summary <= '0';
	 end if;
	 if(htr_summary = '1' and fifo_din(15) = '1')then
		catch_trigger <= fifo_din(11 downto 10) & not fifo_din(13 downto 12) & fifo_din(31) & fifo_din(26 downto 24);
	 else
		catch_trigger <= x"00";
	 end if;
  end if;
end process;       
--scale(31 downto 16) <= x"0000";
--r_data <= scale when rd_scale = '1' else (others => 'Z');
r_data(31 downto 16) <= scale(31 downto 16) when rd_scale = '1' else "ZZZZZZZZZZZZZZZZ";
r_data(15 downto 0) <= "ZZZZZZZZZZZZZZZZ"  when rd_scale = '0' else scale(15 downto 0) when catch_enable = '0' else catch_offset_i & wc_reg_wa(6 downto 0) & '0' & catch_cntr;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   ce_scale <= '0';
	   ld_scale <= '0';
	   up_count <= '0';
	   ce_wc_reg_wa <= '0';
	   inc_mon_cntr <= '0';
	 else
	   ce_scale <= we_fifo and fifo_din(33);
	   ld_scale <= (mon_en or pending) and we_fifo and fifo_din(33);
	   up_count <= mon_en and we_fifo and fifo_din(33);
	   ce_wc_reg_wa <= mon_we_i and fifo_dout(32);
		if(strobe = '1')then
	     inc_mon_cntr <= ce_wc_reg_wa or (mon_we_i and fifo_dout(32));
		end if;
	 end if;
  end if;
end process;
not_scale <= not scale(15 downto 0);
i_mon_en: l_cntr
  generic map(N => 16)
  port map(
    d => not_scale,
    load => ld_scale,
    reset => '0',
    set => '0',
    ec => ce_scale,
    en_cnt => '1',
    clk => clk,
	 carry => scale_carry);
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
      mon_enp2 <= '0';
    elsif(slink_we = '1' and slink_data(33) = '1')then
		mon_enp2 <= slink_data(35) and not catch_cntr(6) and not evt_cnt(7);
	 end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
-- After reset, the first event will always be recorded
      mon_enp <= '1';
	   pending <= '0';
    elsif(we_fifo = '1' and fifo_din(32) = '1')then
		if(catch_enable = '1')then
		  mon_enp <= not catch_cntr(6);
		  pending <= '0';
		else
		  mon_enp <= not evt_cnt(7) and (scale_carry or pending);
		  pending <= evt_cnt(7) and (scale_carry or pending);
		end if;
	 end if;
  end if;
end process;
--mon_en <= mon_enp or (fifo_din(35) and not catch_cntr(6) and not evt_cnt(7));
mon_en <= mon_enp or mon_enp2;
end my_arch;
